@@ -8,9 +8,10 @@ Various constants
 """
 
 from collections import defaultdict
-from enum import Enum
-from six import PY2
 from collections import namedtuple
+from enum import Enum
+
+from six import PY2
 
 DBAPI_TYPE_STRING = 0
 DBAPI_TYPE_BINARY = 1
@@ -126,3 +127,20 @@ FileHeader = namedtuple(
         "encryption_metadata"
     ]
 )
+
+PARAMETER_AUTOCOMMIT = u'AUTOCOMMIT'
+PARAMETER_CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY = u'CLIENT_SESSION_KEEP_ALIVE_HEARTBEAT_FREQUENCY'
+PARAMETER_CLIENT_SESSION_KEEP_ALIVE = u'CLIENT_SESSION_KEEP_ALIVE'
+PARAMETER_CLIENT_TELEMETRY_ENABLED = u'CLIENT_TELEMETRY_ENABLED'
+PARAMETER_CLIENT_STORE_TEMPORARY_CREDENTIAL = u'CLIENT_STORE_TEMPORARY_CREDENTIAL'
+PARAMETER_CLIENT_USE_SECURE_STORAGE_FOR_TEMPORARY_CREDENTIAL = \
+    u'CLIENT_USE_SECURE_STORAGE_FOR_TEMPORARY_CREDENTAIL'
+PARAMETER_TIMEZONE = u'TIMEZONE'
+PARAMETER_SERVICE_NAME = u'SERVICE_NAME'
+
+HTTP_HEADER_CONTENT_TYPE = u'Content-Type'
+HTTP_HEADER_ACCEPT = u"accept"
+HTTP_HEADER_USER_AGENT = u"User-Agent"
+HTTP_HEADER_SERVICE_NAME = u'X-Snowflake-Service'
+
+HTTP_HEADER_VALUE_OCTET_STREAM = u'application/octet-stream'
