@@ -10,8 +10,6 @@ import itertools
 import json
 import logging
 import platform
-import random
-import ssl
 import sys
 import time
 import uuid
@@ -26,36 +24,43 @@ from botocore.vendored.requests.exceptions import (
     ConnectionError, ConnectTimeout, ReadTimeout, SSLError)
 from botocore.vendored.requests.packages.urllib3.exceptions import (
     ProtocolError, ReadTimeoutError)
-from botocore.vendored.requests.packages.urllib3.exceptions \
-    import SSLError as urllib3_SSLError
 
-from . import proxy
+from snowflake.connector.time_util import get_time_millis
 from . import ssl_wrap_socket
 from .compat import (
-    BAD_REQUEST, SERVICE_UNAVAILABLE, GATEWAY_TIMEOUT,
+    PY2,
+    METHOD_NOT_ALLOWED, BAD_REQUEST, SERVICE_UNAVAILABLE, GATEWAY_TIMEOUT,
     FORBIDDEN, BAD_GATEWAY, REQUEST_TIMEOUT,
     UNAUTHORIZED, INTERNAL_SERVER_ERROR, OK, BadStatusLine)
-from .compat import (TO_UNICODE, urlencode)
-from .compat import proxy_bypass
+from .compat import (TO_UNICODE, urlencode, urlparse)
+from .constants import (
+    HTTP_HEADER_CONTENT_TYPE,
+    HTTP_HEADER_ACCEPT,
+    HTTP_HEADER_USER_AGENT,
+    HTTP_HEADER_SERVICE_NAME,
+)
 from .errorcode import (ER_FAILED_TO_CONNECT_TO_DB, ER_CONNECTION_IS_CLOSED,
-                        ER_FAILED_TO_REQUEST, ER_FAILED_TO_RENEW_SESSION,
-                        ER_INVALID_CERTIFICATE)
+                        ER_FAILED_TO_REQUEST, ER_FAILED_TO_RENEW_SESSION)
 from .errors import (Error, OperationalError, DatabaseError, ProgrammingError,
                      GatewayTimeoutError, ServiceUnavailableError,
                      InterfaceError, InternalServerError, ForbiddenError,
-                     BadGatewayError, BadRequest,
-                     OtherHTTPRetryableError,
-                     ER_MSG_FAILED_TO_VALIDATE_SSL_CERTIFICATE,
-                     ER_MSG_FAILED_TO_VALIDATE_SSL_CERTIFICATE_SNOWSQL)
+                     BadGatewayError, BadRequest, MethodNotAllowed,
+                     OtherHTTPRetryableError)
 from .gzip_decoder import decompress_raw_data
 from .sqlstate import (SQLSTATE_CONNECTION_NOT_EXISTS,
                        SQLSTATE_CONNECTION_WAS_NOT_ESTABLISHED,
                        SQLSTATE_CONNECTION_REJECTED)
-from .ssl_wrap_socket import (set_proxies)
+from .time_util import (
+    DecorrelateJitterBackoff,
+    DEFAULT_MASTER_VALIDITY_IN_SECONDS
+)
 from .tool.probe_connection import probe_connection
 from .util_text import split_rows_from_stream
 from .version import VERSION
 
+if PY2:
+    from pyasn1.error import PyAsn1Error
+
 logger = logging.getLogger(__name__)
 
 """
@@ -73,21 +78,33 @@ DEFAULT_SOCKET_CONNECT_TIMEOUT = 1 * 60  # don't reduce less than 45 seconds
 # return codes
 QUERY_IN_PROGRESS_CODE = u'333333'  # GS code: the query is in progress
 QUERY_IN_PROGRESS_ASYNC_CODE = u'333334'  # GS code: the query is detached
+
+ID_TOKEN_EXPIRED_GS_CODE = u'390110'
 SESSION_EXPIRED_GS_CODE = u'390112'  # GS code: session expired. need to renew
+MASTER_TOKEN_NOTFOUND_GS_CODE = u'390113'
+MASTER_TOKEN_EXPIRED_GS_CODE = u'390114'
+MASTER_TOKEN_INVALD_GS_CODE = u'390115'
+BAD_REQUEST_GS_CODE = u'390400'
 
 # other constants
 CONTENT_TYPE_APPLICATION_JSON = u'application/json'
 ACCEPT_TYPE_APPLICATION_SNOWFLAKE = u'application/snowflake'
 
 REQUEST_TYPE_RENEW = u'RENEW'
-REQUEST_TYPE_CLONE = u'CLONE'
 REQUEST_TYPE_ISSUE = u'ISSUE'
 
+UPDATED_BY_ID_TOKEN = u'updated_by_id_token'
+
 HEADER_AUTHORIZATION_KEY = u"Authorization"
 HEADER_SNOWFLAKE_TOKEN = u'Snowflake Token="{token}"'
 
+REQUEST_ID = u'requestId'
+REQUEST_GUID = u'request_guid'
+SNOWFLAKE_HOST_SUFFIX = u'.snowflakecomputing.com'
+
 SNOWFLAKE_CONNECTOR_VERSION = u'.'.join(TO_UNICODE(v) for v in VERSION[0:3])
 PYTHON_VERSION = u'.'.join(TO_UNICODE(v) for v in sys.version_info[:3])
+OPERATING_SYSTEM = platform.system()
 PLATFORM = platform.platform()
 IMPLEMENTATION = platform.python_implementation()
 COMPILER = platform.python_compiler()
@@ -110,16 +127,23 @@ STATUS_TO_EXCEPTION = {
     GATEWAY_TIMEOUT: GatewayTimeoutError,
     BAD_REQUEST: BadRequest,
     BAD_GATEWAY: BadGatewayError,
+    METHOD_NOT_ALLOWED: MethodNotAllowed,
 }
 
+DEFAULT_AUTHENTICATOR = u'SNOWFLAKE'  # default authenticator name
+EXTERNAL_BROWSER_AUTHENTICATOR = u'EXTERNALBROWSER'
+KEY_PAIR_AUTHENTICATOR = u'SNOWFLAKE_JWT'
+OAUTH_AUTHENTICATOR = u'OAUTH'
+
 
 def is_retryable_http_code(code):
     """
     Is retryable HTTP code?
     """
-    return code >= 500 and code < 600 or code in (
+    return 500 <= code < 600 or code in (
         BAD_REQUEST,  # 400
         FORBIDDEN,  # 403
+        METHOD_NOT_ALLOWED,  # 405
         REQUEST_TIMEOUT,  # 408
     )
 
@@ -131,6 +155,15 @@ class RetryRequest(Exception):
     pass
 
 
+class ReauthenticationRequest(Exception):
+    """
+    Signal to reauthenticate
+    """
+
+    def __init__(self, cause):
+        self.cause = cause
+
+
 class SnowflakeAuth(AuthBase):
     """
     Attaches HTTP Authorization header for Snowflake
@@ -151,25 +184,35 @@ class SnowflakeAuth(AuthBase):
         return r
 
 
+class ResultIterWithTimings(collections.Iterator):
+    DOWNLOAD = u"download"
+    PARSE = u"parse"
+
+    def __init__(self, it, timings):
+        self._it = it
+        self._timings = timings
+
+    def __next__(self):
+        return next(self._it)
+
+    def next(self):
+        return self.__next__()
+
+    def get_timings(self):
+        return self._timings
+
+
 class SnowflakeRestful(object):
     """
     Snowflake Restful class
     """
 
     def __init__(self, host=u'127.0.0.1', port=8080,
-                 proxy_host=None,
-                 proxy_port=None,
-                 proxy_user=None,
-                 proxy_password=None,
                  protocol=u'http',
                  inject_client_pause=0,
                  connection=None):
         self._host = host
         self._port = port
-        self._proxy_host = proxy_host
-        self._proxy_port = proxy_port
-        self._proxy_user = proxy_user
-        self._proxy_password = proxy_password
         self._protocol = protocol
         self._inject_client_pause = inject_client_pause
         self._connection = connection
@@ -184,14 +227,9 @@ class SnowflakeRestful(object):
         ssl_wrap_socket.FEATURE_OCSP_RESPONSE_CACHE_FILE_NAME = \
             self._connection and self._connection._ocsp_response_cache_filename
         #
-        proxy.PROXY_HOST = self._proxy_host
-        proxy.PROXY_PORT = self._proxy_port
-        proxy.PROXY_USER = self._proxy_user
-        proxy.PROXY_PASSWORD = self._proxy_password
 
         # This is to address the issue where requests hangs
         _ = 'dummy'.encode('idna').decode('utf-8')  # noqa
-        proxy_bypass('www.snowflake.net:443')
 
     @property
     def token(self):
@@ -201,14 +239,43 @@ class SnowflakeRestful(object):
     def master_token(self):
         return self._master_token if hasattr(self, u'_master_token') else None
 
+    @property
+    def master_validity_in_seconds(self):
+        return self._master_validity_in_seconds \
+            if hasattr(self, u'_master_validity_in_seconds') and \
+               self._master_validity_in_seconds \
+            else DEFAULT_MASTER_VALIDITY_IN_SECONDS
+
+    @master_validity_in_seconds.setter
+    def master_validity_in_seconds(self, value):
+        self._master_validity_in_seconds = value \
+            if value else DEFAULT_MASTER_VALIDITY_IN_SECONDS
+
+    @property
+    def id_token(self):
+        return self._id_token if hasattr(self, u'_id_token') else None
+
+    @id_token.setter
+    def id_token(self, value):
+        self._id_token = value
+
+    @property
+    def id_token_password(self):
+        return self._id_token_password if hasattr(
+            self, u'_id_token_password') else None
+
     def close(self):
         if hasattr(self, u'_token'):
             del self._token
         if hasattr(self, u'_master_token'):
             del self._master_token
+        if hasattr(self, u'_id_token'):
+            del self._id_token
+        if hasattr(self, u'_id_token_password'):
+            del self._id_token_password
         sessions = list(self._active_sessions)
         if sessions:
-            logger.warning("Closing %s active sessions", len(sessions))
+            logger.debug("Closing %s active sessions", len(sessions))
         sessions.extend(self._idle_sessions)
         self._active_sessions.clear()
         self._idle_sessions.clear()
@@ -216,13 +283,13 @@ class SnowflakeRestful(object):
             try:
                 s.close()
             except Exception as e:
-                logger.warning("Session cleanup failed: %s", e)
+                logger.info("Session cleanup failed: %s", e)
 
     def request(self, url, body=None, method=u'post', client=u'sfsql',
-                _no_results=False):
+                _no_results=False, timeout=None, _include_retry_params=False):
         if body is None:
             body = {}
-        if self.master_token is None:
+        if self.master_token is None and self.token is None:
             Error.errorhandler_wrapper(
                 self._connection, None, DatabaseError,
                 {
@@ -236,78 +303,150 @@ class SnowflakeRestful(object):
         else:
             accept_type = CONTENT_TYPE_APPLICATION_JSON
 
+        if timeout is None:
+            timeout = self._connection.network_timeout
+
         headers = {
-            u'Content-Type': CONTENT_TYPE_APPLICATION_JSON,
-            u"accept": accept_type,
-            u"User-Agent": PYTHON_CONNECTOR_USER_AGENT,
+            HTTP_HEADER_CONTENT_TYPE: CONTENT_TYPE_APPLICATION_JSON,
+            HTTP_HEADER_ACCEPT: accept_type,
+            HTTP_HEADER_USER_AGENT: PYTHON_CONNECTOR_USER_AGENT,
         }
+        if self._connection.service_name:
+            headers[HTTP_HEADER_SERVICE_NAME] = self._connection.service_name
         if method == u'post':
             return self._post_request(
                 url, headers, json.dumps(body),
-                token=self._token, _no_results=_no_results,
-                timeout=self._connection.network_timeout)
+                token=self.token, _no_results=_no_results,
+                timeout=timeout, _include_retry_params=_include_retry_params)
         else:
             return self._get_request(
-                url, headers, token=self._token,
-                timeout=self._connection.network_timeout)
+                url, headers, token=self.token,
+                timeout=timeout)
 
-    def update_tokens(self, session_token, master_token):
+    def update_tokens(self, session_token, master_token,
+                      master_validity_in_seconds=None,
+                      id_token=None, id_token_password=None):
         """
-        Update session and master tokens
+        Update session and master tokens and optionally temporary credential
         """
         with self._lock_token:
             self._token = session_token
             self._master_token = master_token
+            self._id_token = id_token
+            self._id_token_password = id_token_password
+            self._master_validity_in_seconds = master_validity_in_seconds
 
     def _renew_session(self):
-        if self.master_token is None:
-            Error.errorhandler_wrapper(
-                self._connection, None, DatabaseError,
-                {
-                    u'msg': u"Connection is closed",
-                    u'errno': ER_CONNECTION_IS_CLOSED,
-                    u'sqlstate': SQLSTATE_CONNECTION_NOT_EXISTS,
-                })
+        """
+        Renew a session and master token.
+        """
+        try:
+            return self._token_request(REQUEST_TYPE_RENEW)
+        except ReauthenticationRequest as ex:
+            if not self.id_token:
+                raise ex.cause
+            return self._token_request(REQUEST_TYPE_ISSUE)
 
-        logger.debug(u'updating session. master_token: %s', self.master_token)
+    def _id_token_session(self):
+        """
+        Issue a session token by the id token. No master token is returned.
+        As a result, the session token is not renewable.
+        """
+        return self._token_request(REQUEST_TYPE_ISSUE)
+
+    def _token_request(self, request_type):
+        logger.debug(
+            u'updating session. master_token: %s, id_token: %s',
+            u'****' if self.master_token else None,
+            u'****' if self.id_token else None)
         headers = {
-            u'Content-Type': CONTENT_TYPE_APPLICATION_JSON,
-            u"accept": CONTENT_TYPE_APPLICATION_JSON,
-            u"User-Agent": PYTHON_CONNECTOR_USER_AGENT,
+            HTTP_HEADER_CONTENT_TYPE: CONTENT_TYPE_APPLICATION_JSON,
+            HTTP_HEADER_ACCEPT: CONTENT_TYPE_APPLICATION_JSON,
+            HTTP_HEADER_USER_AGENT: PYTHON_CONNECTOR_USER_AGENT,
         }
+        if self._connection.service_name:
+            headers[HTTP_HEADER_SERVICE_NAME] = self._connection.service_name
         request_id = TO_UNICODE(uuid.uuid4())
         logger.debug(u'request_id: %s', request_id)
         url = u'/session/token-request?' + urlencode({
-            u'requestId': request_id})
+            REQUEST_ID: request_id})
 
-        body = {
-            u"oldSessionToken": self._token,
-            u"requestType": REQUEST_TYPE_RENEW,
-        }
+        if request_type == REQUEST_TYPE_ISSUE:
+            header_token = self.id_token
+            body = {
+                u"idToken": self.id_token,
+                u"requestType": REQUEST_TYPE_ISSUE,
+            }
+        else:
+            # NOTE: ensure an empty key if master token is not set.
+            # This avoids HTTP 400.
+            header_token = self.master_token or ""
+            body = {
+                u"oldSessionToken": self.token,
+                u"requestType": request_type,
+            }
         ret = self._post_request(
             url, headers, json.dumps(body),
-            token=self.master_token,
+            token=header_token,
             timeout=self._connection.network_timeout)
-        if ret.get(u'success') and u'data' in ret \
-                and u'sessionToken' in ret[u'data']:
+        if ret.get(u'success') and ret.get(u'data', {}).get(u'sessionToken'):
             logger.debug(u'success: %s', ret)
             self.update_tokens(
-                ret[u'data'][u'sessionToken'], ret[u'data'][u'masterToken'])
+                ret[u'data'][u'sessionToken'],
+                ret[u'data'].get(u'masterToken'),
+                master_validity_in_seconds=ret[u'data'].get(
+                    u'masterValidityInSeconds'),
+                id_token=self.id_token,
+                id_token_password=self.id_token_password)
             logger.debug(u'updating session completed')
+            ret[UPDATED_BY_ID_TOKEN] = request_type == REQUEST_TYPE_ISSUE
             return ret
         else:
             logger.debug(u'failed: %s', ret)
             err = ret.get(u'message')
             if err is not None and ret.get(u'data'):
                 err += ret[u'data'].get(u'errorMessage', '')
+            errno = ret.get(u'code') or ER_FAILED_TO_RENEW_SESSION
+            if errno in (
+                    ID_TOKEN_EXPIRED_GS_CODE,
+                    SESSION_EXPIRED_GS_CODE,
+                    MASTER_TOKEN_NOTFOUND_GS_CODE,
+                    MASTER_TOKEN_EXPIRED_GS_CODE,
+                    MASTER_TOKEN_INVALD_GS_CODE,
+                    BAD_REQUEST_GS_CODE):
+                raise ReauthenticationRequest(
+                    ProgrammingError(
+                        msg=err,
+                        errno=int(errno),
+                        sqlstate=SQLSTATE_CONNECTION_WAS_NOT_ESTABLISHED))
             Error.errorhandler_wrapper(
                 self._connection, None, ProgrammingError,
                 {
                     u'msg': err,
-                    u'errno': ER_FAILED_TO_RENEW_SESSION,
+                    u'errno': int(errno),
                     u'sqlstate': SQLSTATE_CONNECTION_WAS_NOT_ESTABLISHED,
                 })
 
+    def _heartbeat(self):
+        headers = {
+            HTTP_HEADER_CONTENT_TYPE: CONTENT_TYPE_APPLICATION_JSON,
+            HTTP_HEADER_ACCEPT: CONTENT_TYPE_APPLICATION_JSON,
+            HTTP_HEADER_USER_AGENT: PYTHON_CONNECTOR_USER_AGENT,
+        }
+        if self._connection.service_name:
+            headers[HTTP_HEADER_SERVICE_NAME] = self._connection.service_name
+        request_id = TO_UNICODE(uuid.uuid4())
+        logger.debug(u'request_id: %s', request_id)
+        url = u'/session/heartbeat?' + urlencode({
+            REQUEST_ID: request_id})
+        ret = self._post_request(
+            url, headers, None,
+            token=self.token,
+            timeout=self._connection.network_timeout)
+        if not ret.get(u'success'):
+            logger.error("Failed to heartbeat. code: %s, url: %s",
+                         ret.get(u'code'), url)
+
     def delete_session(self):
         """
         Deletes the session
@@ -323,15 +462,18 @@ class SnowflakeRestful(object):
 
         url = u'/session?' + urlencode({u'delete': u'true'})
         headers = {
-            u'Content-Type': CONTENT_TYPE_APPLICATION_JSON,
-            u"accept": CONTENT_TYPE_APPLICATION_JSON,
-            u"User-Agent": PYTHON_CONNECTOR_USER_AGENT,
+            HTTP_HEADER_CONTENT_TYPE: CONTENT_TYPE_APPLICATION_JSON,
+            HTTP_HEADER_ACCEPT: CONTENT_TYPE_APPLICATION_JSON,
+            HTTP_HEADER_USER_AGENT: PYTHON_CONNECTOR_USER_AGENT,
         }
+        if self._connection.service_name:
+            headers[HTTP_HEADER_SERVICE_NAME] = self._connection.service_name
+
         body = {}
         try:
             ret = self._post_request(
                 url, headers, json.dumps(body),
-                token=self._token, timeout=5, no_retry=True)
+                token=self.token, timeout=5, no_retry=True)
             if not ret or ret.get(u'success'):
                 return
             err = ret.get(u'message')
@@ -359,18 +501,27 @@ class SnowflakeRestful(object):
         ret = self.fetch(u'get', full_url, headers, timeout=timeout,
                          token=token, socket_timeout=socket_timeout)
         if ret.get(u'code') == SESSION_EXPIRED_GS_CODE:
-            ret = self._renew_session()
+            try:
+                ret = self._renew_session()
+                if ret.get(UPDATED_BY_ID_TOKEN):
+                    self._connection._set_current_objects()
+            except ReauthenticationRequest as ex:
+                if self._connection._authenticator != \
+                        EXTERNAL_BROWSER_AUTHENTICATOR:
+                    raise ex.cause
+                ret = self._connection._reauthenticate_by_webbrowser()
             logger.debug(
                 u'ret[code] = {code} after renew_session'.format(
                     code=(ret.get(u'code', u'N/A'))))
             if ret.get(u'success'):
-                return self._get_request(url, headers, token=self._token)
+                return self._get_request(url, headers, token=self.token)
 
         return ret
 
     def _post_request(self, url, headers, body, token=None,
                       timeout=None, _no_results=False, no_retry=False,
-                      socket_timeout=DEFAULT_SOCKET_CONNECT_TIMEOUT):
+                      socket_timeout=DEFAULT_SOCKET_CONNECT_TIMEOUT,
+                      _include_retry_params=False):
         full_url = u'{protocol}://{host}:{port}{url}'.format(
             protocol=self._protocol,
             host=self._host,
@@ -384,19 +535,28 @@ class SnowflakeRestful(object):
 
         ret = self.fetch(u'post', full_url, headers, data=body,
                          timeout=timeout, token=token,
-                         no_retry=no_retry, socket_timeout=socket_timeout)
+                         no_retry=no_retry, socket_timeout=socket_timeout,
+                         _include_retry_params=_include_retry_params)
         logger.debug(
             u'ret[code] = {code}, after post request'.format(
                 code=(ret.get(u'code', u'N/A'))))
 
         if ret.get(u'code') == SESSION_EXPIRED_GS_CODE:
-            ret = self._renew_session()
+            try:
+                ret = self._renew_session()
+                if ret.get(UPDATED_BY_ID_TOKEN):
+                    self._connection._set_current_objects()
+            except ReauthenticationRequest as ex:
+                if self._connection._authenticator != \
+                        EXTERNAL_BROWSER_AUTHENTICATOR:
+                    raise ex.cause
+                ret = self._connection._reauthenticate_by_webbrowser()
             logger.debug(
                 u'ret[code] = {code} after renew_session'.format(
-                    code=(ret.get(u'code'u'N/A'))))
+                    code=(ret.get(u'code', u'N/A'))))
             if ret.get(u'success'):
                 return self._post_request(
-                    url, headers, body, token=self._token, timeout=timeout)
+                    url, headers, body, token=self.token, timeout=timeout)
 
         if ret.get(u'code') == QUERY_IN_PROGRESS_ASYNC_CODE and _no_results:
             return ret
@@ -405,14 +565,13 @@ class SnowflakeRestful(object):
                 (QUERY_IN_PROGRESS_CODE, QUERY_IN_PROGRESS_ASYNC_CODE):
             if self._inject_client_pause > 0:
                 logger.debug(
-                    u'waiting for {inject_client_pause}...'.format(
-                        inject_client_pause=self._inject_client_pause))
+                    u'waiting for %s...', self._inject_client_pause)
                 time.sleep(self._inject_client_pause)
             # ping pong
             result_url = ret[u'data'][u'getResultUrl']
             logger.debug(u'ping pong starting...')
             ret = self._get_request(
-                result_url, headers, token=self._token, timeout=timeout)
+                result_url, headers, token=self.token, timeout=timeout)
             logger.debug(u'ret[code] = %s', ret.get(u'code', u'N/A'))
             logger.debug(u'ping pong done')
 
@@ -422,22 +581,13 @@ class SnowflakeRestful(object):
               **kwargs):
         """ Curried API request with session management. """
 
-        class DecorrelateJitterBackoff(object):
-            # Decorrelate Jitter backoff
-            # see:
-            # https://www.awsarchitectureblog.com/2015/03/backoff.html
-            def __init__(self, base, cap):
-                self._base = base
-                self._cap = cap
-
-            def next_sleep(self, _, sleep):
-                return min(self._cap, random.randint(self._base, sleep * 3))
-
         class RetryCtx(object):
-            def __init__(self, timeout):
+            def __init__(self, timeout, _include_retry_params=False):
                 self.timeout = timeout
                 self.cnt = 0
                 self.sleeping_time = 1
+                self.start_time = get_time_millis()
+                self._include_retry_params = _include_retry_params
                 # backoff between 1 and 16 seconds
                 self._backoff = DecorrelateJitterBackoff(1, 16)
 
@@ -446,8 +596,21 @@ class SnowflakeRestful(object):
                     self.cnt, self.sleeping_time)
                 return self.sleeping_time
 
+            def add_retry_params(self, full_url):
+                if self._include_retry_params and self.cnt > 0:
+                    suffix = urlencode({
+                        'clientStartTime': self.start_time,
+                        'retryCount': self.cnt
+                    })
+                    sep = '&' if urlparse(full_url).query else '?'
+                    return full_url + sep + suffix
+                else:
+                    return full_url
+
+        include_retry_params = kwargs.pop('_include_retry_params', False)
+
         with self._use_requests_session() as session:
-            retry_ctx = RetryCtx(timeout)
+            retry_ctx = RetryCtx(timeout, include_retry_params)
             while True:
                 ret = self._request_exec_wrapper(
                     session, method, full_url, headers, data, retry_ctx,
@@ -455,6 +618,21 @@ class SnowflakeRestful(object):
                 if ret is not None:
                     return ret
 
+    @staticmethod
+    def add_request_guid(full_url):
+        """
+        Add request_guid parameter for HTTP request tracing
+        """
+        parsed_url = urlparse(full_url)
+        if not parsed_url.hostname.endswith(SNOWFLAKE_HOST_SUFFIX):
+            return full_url
+        suffix = urlencode({
+            REQUEST_GUID: TO_UNICODE(uuid.uuid4())
+        })
+        sep = '&' if parsed_url.query else '?'
+        # url has query string already, just add fields
+        return full_url + sep + suffix
+
     def _request_exec_wrapper(
             self,
             session, method, full_url, headers, data, retry_ctx,
@@ -462,12 +640,12 @@ class SnowflakeRestful(object):
             **kwargs):
 
         conn = self._connection
-        proxies = set_proxies(self._proxy_host, self._proxy_port,
-                              self._proxy_user, self._proxy_password)
         logger.debug('remaining request timeout: %s, retry cnt: %s',
                      retry_ctx.timeout, retry_ctx.cnt + 1)
 
         start_request_thread = time.time()
+        full_url = retry_ctx.add_retry_params(full_url)
+        full_url = SnowflakeRestful.add_request_guid(full_url)
         try:
             return_object = self._request_exec(
                 session=session,
@@ -475,13 +653,12 @@ class SnowflakeRestful(object):
                 full_url=full_url,
                 headers=headers,
                 data=data,
-                proxies=proxies,
                 token=token,
                 **kwargs)
             if return_object is not None:
                 return return_object
             self._handle_unknown_error(
-                method, full_url, headers, data, conn, proxies)
+                method, full_url, headers, data, conn)
             return {}
         except RetryRequest as e:
             if no_retry:
@@ -508,9 +685,9 @@ class SnowflakeRestful(object):
                 retry_ctx.cnt + 1,
                 sleeping_time)
             time.sleep(sleeping_time)
+            retry_ctx.cnt += 1
             if retry_ctx.timeout is not None:
                 retry_ctx.timeout -= sleeping_time
-                retry_ctx.cnt += 1
             return None  # retry
         except Exception as e:
             if not no_retry:
@@ -519,29 +696,6 @@ class SnowflakeRestful(object):
             return {}
 
     def handle_invalid_certificate_error(self, conn, full_url, cause):
-        # checking if the cause is SSL certificate error
-        if isinstance(cause, SSLError):
-            if isinstance(cause.args[0], urllib3_SSLError):
-                if isinstance(
-                        cause.args[0].args[0], ssl.SSLError):
-                    err = cause.args[0].args[0].args[0]
-                    if u'bad handshake' in err and \
-                            u'certificate verify failed' in err:
-                        if conn.application == APPLICATION_SNOWSQL:
-                            # SnowSQL has an option to probe connection.
-                            msg = \
-                                ER_MSG_FAILED_TO_VALIDATE_SSL_CERTIFICATE_SNOWSQL
-                        else:
-                            msg = ER_MSG_FAILED_TO_VALIDATE_SSL_CERTIFICATE
-                        Error.errorhandler_wrapper(
-                            conn, None, OperationalError,
-                            {
-                                u'errno': ER_INVALID_CERTIFICATE,
-                                u'msg': msg.format(full_url),
-                                u'error': ER_FAILED_TO_REQUEST,
-                            }
-                        )
-                        return  # required for tests
         # all other errors raise exception
         Error.errorhandler_wrapper(
             conn, None, OperationalError,
@@ -552,7 +706,7 @@ class SnowflakeRestful(object):
             })
 
     def _handle_unknown_error(
-            self, method, full_url, headers, data, conn, proxies):
+            self, method, full_url, headers, data, conn):
         """
         Handle unknown error
         """
@@ -569,23 +723,20 @@ class SnowflakeRestful(object):
         logger.error(
             u'Failed to get the response. Hanging? '
             u'method: {method}, url: {url}, headers:{headers}, '
-            u'data: {data}, proxies: {proxies}'.format(
+            u'data: {data}'.format(
                 method=method,
                 url=full_url,
                 headers=headers,
                 data=data,
-                proxies=proxies
             )
         )
         Error.errorhandler_wrapper(
             conn, None, OperationalError,
             {
                 u'msg': u'Failed to get the response. Hanging? '
-                        u'method: {method}, url: {url}, '
-                        u'proxies: {proxies}'.format(
+                        u'method: {method}, url: {url}'.format(
                     method=method,
                     url=full_url,
-                    proxies=proxies
                 ),
                 u'errno': ER_FAILED_TO_REQUEST,
             })
@@ -593,14 +744,14 @@ class SnowflakeRestful(object):
     def _request_exec(
             self,
             session, method, full_url, headers, data,
-            proxies,
             token,
             catch_okta_unauthorized_error=False,
             is_raw_text=False,
             is_raw_binary=False,
             is_raw_binary_iterator=True,
             use_ijson=False,
-            socket_timeout=DEFAULT_SOCKET_CONNECT_TIMEOUT):
+            socket_timeout=DEFAULT_SOCKET_CONNECT_TIMEOUT,
+            return_timing_metrics=False):
         if socket_timeout > DEFAULT_SOCKET_CONNECT_TIMEOUT:
             # socket timeout should not be more than the default.
             # A shorter timeout may be specified for login time, but
@@ -618,13 +769,14 @@ class SnowflakeRestful(object):
             else:
                 input_data = data
 
+            timing_metrics = {}
+            start_time = get_time_millis()
             # socket timeout is constant. You should be able to receive
             # the response within the time. If not, ConnectReadTimeout or
             # ReadTimeout is raised.
             raw_ret = session.request(
                 method=method,
                 url=full_url,
-                proxies=proxies,
                 headers=headers,
                 data=input_data,
                 timeout=socket_timeout,
@@ -632,12 +784,16 @@ class SnowflakeRestful(object):
                 stream=is_raw_binary,
                 auth=SnowflakeAuth(token),
             )
+            timing_metrics[
+                ResultIterWithTimings.DOWNLOAD] = get_time_millis() - start_time
+
             try:
                 if raw_ret.status_code == OK:
                     logger.debug(u'SUCCESS')
                     if is_raw_text:
                         ret = raw_ret.text
                     elif is_raw_binary:
+                        start_time = get_time_millis()
                         raw_data = decompress_raw_data(
                             raw_ret.raw, add_bracket=True
                         ).decode('utf-8', 'replace')
@@ -647,12 +803,18 @@ class SnowflakeRestful(object):
                             ret = iter(json.loads(raw_data))
                         else:
                             ret = split_rows_from_stream(StringIO(raw_data))
+                        timing_metrics[
+                            ResultIterWithTimings.PARSE] = get_time_millis() - start_time
+
+                        # if timings requested, wrap the iterator so the timing
+                        # info is accessible
+                        if return_timing_metrics:
+                            ret = ResultIterWithTimings(ret, timing_metrics)
                     else:
                         ret = raw_ret.json()
                     return ret
 
-                if is_retryable_http_code(
-                        raw_ret.status_code):
+                if is_retryable_http_code(raw_ret.status_code):
                     ex = STATUS_TO_EXCEPTION.get(
                         raw_ret.status_code, OtherHTTPRetryableError)
                     exi = ex(code=raw_ret.status_code)
@@ -710,6 +872,14 @@ class SnowflakeRestful(object):
                 "error stack: %s", err,
                 exc_info=True)
             raise RetryRequest(err)
+        except Exception as err:
+            if PY2 and isinstance(err, PyAsn1Error):
+                logger.debug(
+                    "Hit retryable client error. Retrying... "
+                    "Ignore the following error stack: %s", err,
+                    exc_info=True)
+                raise RetryRequest(err)
+            raise err
 
     def make_requests_session(self):
         s = requests.Session()
