@@ -21,6 +21,7 @@ FEATURE_OCSP_RESPONSE_CACHE_FILE_NAME = None
 import logging
 import ssl
 import sys
+import time
 from socket import error as SocketError
 from socket import (socket, timeout)
 
@@ -33,10 +34,9 @@ from cryptography import x509
 from cryptography.hazmat.backends.openssl import backend as openssl_backend
 from cryptography.hazmat.backends.openssl.x509 import _Certificate
 
+from .compat import PY2
 from .errorcode import (ER_SERVER_CERTIFICATE_REVOKED)
 from .errors import (OperationalError)
-from .proxy import (set_proxies, PROXY_HOST, PROXY_PORT, PROXY_USER,
-                    PROXY_PASSWORD)
 from .ssl_wrap_util import wait_for_read, wait_for_write
 
 try:  # Platform-specific: Python 2
@@ -137,7 +137,7 @@ def get_subj_alt_name(peer_cert):
     except x509.ExtensionNotFound:
         # No such extension, return the empty list.
         return []
-    except (x509.DuplicateExtension, x509.UnsupportedExtension,
+    except (x509.DuplicateExtension,
             x509.UnsupportedGeneralNameType, UnicodeError) as e:
         # A problem has been found with the quality of the certificate. Assume
         # no SAN field is present.
@@ -373,19 +373,19 @@ def ssl_wrap_socket_with_ocsp(
     global FEATURE_INSECURE_MODE
     global FEATURE_OCSP_RESPONSE_CACHE_FILE_NAME
 
-    from .ocsp_asn1crypto import SnowflakeOCSP
+    if PY2:
+        # Python 2 uses pyasn1 for workaround. For some reason, asn1crypto
+        # fails to parse OCSP response in Python 2.
+        from .ocsp_pyasn1 import SnowflakeOCSPPyasn1 as SFOCSP
+    else:
+        from .ocsp_asn1crypto import SnowflakeOCSPAsn1Crypto as SFOCSP
 
     log.debug(u'insecure_mode: %s, '
-              u'OCSP response cache file name: %s, '
-              u'PROXY_HOST: %s, PROXY_PORT: %s, PROXY_USER: %s '
-              u'PROXY_PASSWORD: %s',
+              u'OCSP response cache file name: %s',
               FEATURE_INSECURE_MODE,
-              FEATURE_OCSP_RESPONSE_CACHE_FILE_NAME,
-              PROXY_HOST, PROXY_PORT, PROXY_USER, PROXY_PASSWORD)
+              FEATURE_OCSP_RESPONSE_CACHE_FILE_NAME)
     if not FEATURE_INSECURE_MODE:
-        v = SnowflakeOCSP(
-            proxies=set_proxies(
-                PROXY_HOST, PROXY_PORT, PROXY_USER, PROXY_PASSWORD),
+        v = SFOCSP(
             ocsp_response_cache_uri=FEATURE_OCSP_RESPONSE_CACHE_FILE_NAME,
         ).validate(server_hostname, ret.connection)
         if not v:
@@ -404,16 +404,33 @@ def ssl_wrap_socket_with_ocsp(
     return ret
 
 
-def _openssl_connect(hostname, port=443):
+def _openssl_connect(hostname, port=443, max_retry=20):
     """
     OpenSSL connection without validating certificates. This is used to diagnose
     SSL issues.
     """
-    client = socket()
-    client.connect((hostname, port))
-    client_ssl = OpenSSL.SSL.Connection(
-        OpenSSL.SSL.Context(OpenSSL.SSL.SSLv23_METHOD), client)
-    client_ssl.set_connect_state()
-    client_ssl.set_tlsext_host_name(hostname.encode('utf-8'))
-    client_ssl.do_handshake()
-    return client_ssl
+    err = None
+    sleeping_time = 1
+    for retry in range(max_retry):
+        try:
+            client = socket()
+            # client.settimeout(5)
+            client.connect((hostname, port))
+            client_ssl = OpenSSL.SSL.Connection(
+                OpenSSL.SSL.Context(OpenSSL.SSL.SSLv23_METHOD), client)
+            client_ssl.set_connect_state()
+            client_ssl.set_tlsext_host_name(hostname.encode('utf-8'))
+            client_ssl.do_handshake()
+            return client_ssl
+        except Exception as ex:
+            if isinstance(ex, OpenSSL.SSL.SysCallError) or \
+                    ex.__class__.__name__ == 'TimeoutError':
+                err = ex
+                sleeping_time *= 2
+                if sleeping_time > 16:
+                    sleeping_time = 16
+                time.sleep(sleeping_time)
+            else:
+                raise ex
+    if err:
+        raise err
